# -*- coding: utf-8 -*-
"""Module containing loguru log record patchers."""

from typing import Any, Iterable

from loguru._logger import Logger

# An interable of patchers we utilize by default for our root `get_logger()` method
# and our `instance` attribute
DEFAULT_PATCHERS: Iterable[Any] = []


def patch_logger(logger: Logger, patchers: Iterable[Any] = DEFAULT_PATCHERS) -> Logger:
    """Produce a new *wrapping* logger that has been patched with the given patchers.
    .. note:: You will probably never need to utilize this method to manually patch
        A logger instance with a specific patcher. This is simply a helper method
        to auto-apply patches required for the application root logger.
        Likely all you need to use to begin logging is to utilize the provided logger
        instance:
        .. code-block:: python
            from modist.log import instance as log
            log.info("This is a logged message")
    Usage of this method typically looks something like the following:
    :param ~loguru._logger.Logger logger: The root logger instance to patch
    :param Iterable[Any] patchers: A list of patcher modules to use when producing a
        new patched logger instsance, optional, defaults to DEFAULT_PATCHERS
    :return: A new patched logger instance
    :rtype: ~loguru._logger.Logger
    """

    patched_logger: Logger = logger

    for patcher in patchers:
        patched_logger = patched_logger.patch(patcher.patch)

    return patched_logger



# -*- coding: utf-8 -*-
"""This module provides unified access to simple to use but powerful logging.
The supplied logger instance is actually an instance of :class:`~loguru._logger.Logger`
from `Loguru <https://loguru.readthedocs.io>`_. This library allows us to perform
advanced structured logging and render informative traceback information to actually
make our logs *worth the effort*.
>>> from lognub.log import log_instance as log
>>> log.info("Some kind of informational message")
Of course, all the standard logging levels are available...
>>> log.trace
>>> log.debug
>>> log.info
>>> log.warning
>>> log.success
>>> log.error
>>> log.critical
>>> log.exception
Any kind of addtional information that needs to be added to the ``extras`` field of the
resulting log record can be done quickly and easily using the
:meth:`~loguru._logger.Logger.bind` method...
>>> ip_address = user.get_ip()  # hypothetical method
>>> log.bind(ip=ip_address).info(f"User {user.id} has just logged in")
Nifty exception logging can be done for an entire method by using the
:meth:`~loguru._logger.Logger.catch` method (just be sure to stay Pythonic and always
re-raise the caught exception)...
>>> @log.catch(reraise=True)
... def failing_method():
...     return 1 / 0
Other features are also available, please checkout
`the loguru documentation <https://loguru.readthedocs.io>`_.
"""

from lognub.client import configure_logger, get_logger, log, log_instance

__all__ = ["get_logger", "configure_logger", "log_instance", "log"]
__version__ = "0.1.0"



# -*- coding: utf-8 -*-
import logging
from typing import Any, Dict, Optional

from loguru._logger import Logger

from lognub.handles._common import BaseLogHandler


class PropagateHandler(BaseLogHandler):
    """Propagate Loguru's logging records to Python's builtin logging.
    You will probaly want to utilize this handler in order to have your loguru log
    records being logged over Python's builtin logging API for better compatibility with
    external tools that explicitly monitor Python's builtin logging
    (such as Sentry, Prometheus, etc.).
    """

    _handler_reference: Dict[Any, int] = {}

    class LoggingHandler(logging.Handler):
        """Log handler that propagates Loguru logs to Python's builtin logging."""

        def handle(self, record: logging.LogRecord):
            """Given a :class:`logging.LogRecord`, handle it with Python logging.
            :param ~logging.LogRecord record: The log record to handle
            """

            logging.getLogger(record.name).handle(record)

    @staticmethod
    def _get_config() -> Dict[str, Any]:
        """Get the loguru configuration dictionary that should be used for this handler.
        :return: The configuration dictionary to add to loguru's logger, or None
        :rtype: Dict[str, Any]
        """

        return dict(  # noqa: C408
            sink=PropagateHandler.LoggingHandler(level=logging.DEBUG),
            format="{message}",  # noqa: FS003
        )

    @classmethod
    def _get_handler_id(cls, logger: Logger) -> Optional[int]:
        """Get the handler's id for the given logger.
        :param ~loguru._logger.Logger logger: The logger to lookup the handler id for
        :return: The integer id of the given logger's handler, otherwise None
        :rtype: Optional[int]
        """

        if logger in cls._handler_reference:
            return cls._handler_reference[logger]

        return None

    @classmethod
    def is_handled(cls, logger: Logger) -> bool:
        """Quick helper method to check if the given logger is already being handled.
        :param ~loguru._logger.Logger logger: The logger to check if already handled
        :return: True if the logger is already handled, otherwise False
        :rtype: bool
        """

        return cls._get_handler_id(logger) is not None

    @classmethod
    def add_handle(cls, logger: Logger) -> bool:
        """Add the logging handler to the given logger instance.
        :param ~loguru._logger.Logger logger: The logger instance to add the handler to
        :return: True if the handler was added, False if the handle was already present
        :rtype: bool
        """

        if cls.is_handled(logger):
            return False

        handler_id: int = logger.add(**PropagateHandler._get_config())
        cls._handler_reference[logger] = handler_id

        return True

    @classmethod
    def remove_handle(cls, logger: Logger) -> bool:
        """Remove the logging handler from the given logger instance.
        :param ~loguru._logger.Logger logger: The logger instance to remove the handle
        :return: True if the handler was removed, False if there was no handler present
        :rtype: bool
        """

        if not cls.is_handled(logger):
            return False

        handler_id = cls._get_handler_id(logger)
        try:
            if not handler_id:
                return False

            logger.remove(handler_id)
            return True
        except ValueError:
            # NOTE: this occurs in an edge-case where the logger is intialized and a
            # PropagateHandler.Logginghandler is added manually
            # (not using the included methods).
            #
            # This doesn't really cause any issues in normal execution, just that the
            # `handler_id` won't exist in our references. Another reason to only
            # configure the loguru logger only once at startup and avoid doing
            # subsequent calls to `modist.log.configure_logger`.
            return False
        finally:
            if logger in cls._handler_reference:
                del cls._handler_reference[logger]



# -*- coding: utf-8 -*-
import abc

from loguru._logger import Logger


class BaseLogHandler(abc.ABC):
    """The abstract handler interface for how log handlers should be built."""

    @classmethod
    @abc.abstractmethod
    def is_handled(cls, logger: Logger) -> bool:
        """Quick helper method to check if the given logger is already being handled.
        :param ~loguru._logger.Logger logger: The logger to check if already handled
        :return: True if the logger is already handled, otherwise False
        :rtype: bool
        """
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def add_handle(cls, logger: Logger) -> bool:
        """Add the logging handler to the given logger instance.
        :param ~loguru._logger.Logger logger: The logger instance to add the handler to
        :return: True if the handler was added, False if the handle was already present
        :rtype: bool
        """
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def remove_handle(cls, logger: Logger) -> bool:
        """Remove the logging handler from the given logger instance.
        :param ~loguru._logger.Logger logger: The logger instance to remove the handle
        :return: True if the handler was removed, False if there was no handler present
        :rtype: bool
        """
        raise NotImplementedError()



# -*- coding: utf-8 -*-
from lognub.handles._common import BaseLogHandler
from lognub.handles.intercept_handler import InterceptHandler
from lognub.handles.propagate_handler import PropagateHandler

__all__ = ["InterceptHandler", "PropagateHandler", "BaseLogHandler"]



# -*- coding: utf-8 -*-
import copy
import logging
from functools import lru_cache
from typing import Any, List

import loguru
from loguru._logger import Logger

from lognub.handles._common import BaseLogHandler


class InterceptHandler(BaseLogHandler):
    """Intercept Python's builtin logging as Loguru logging records.
    You will probably want to use this when you wish another module's logging to passed
    into whatever Loguru sinks you have configured. This handler was originally added
    for HTTP API frameworks such as werkzeug's logging needing to be passed through into
    whatever sinks are configured by the global loguru logging instance.
    .. important:: This handler intercepts all **already defined** builtin logger logs
        through loguru's log record handling. This means that builtin loggers defined
        **after** this handler is added will not be captured (and will likely break the
        configuration we do to support logging intercepts). So if you need to use
        this handler, do your absolute best to ensure that the call to
        :meth:`~InterceptHandler.add_handle` is performed after other module using
        Python's builtin logging are imported and configured.
    Some example usage of what using this handler looks like is included below:
    >>> log = get_logger()
    >>> from lognub.handles.intercept_handler import InterceptHandler
    >>> InterceptHandler.add_handle(log)
    .. caution:: Python log intercepting is not very robust or dynamic. You bascially
        need to decided if you want to intercept all logs or if you want to ignore other
        module's log's at the very start of runtime. You can't update constructed
        loggers from Python's builtin :mod:`logging` after they have been defined.
        Also any third-party module that calls :func:`logging.basicConfig` and resets
        global logging handlers will break this handler. Your best course of action is
        to just accept that Python's logging will always suck and only worry about your
        own modules logging.
    """

    _is_intercepting: bool = False
    _previous_handlers: List[Any] = []

    class LoggingHandler(logging.Handler):
        """Log handler that intercepts Python's builtin logging for Loguru logs."""

        @lru_cache(maxsize=64, typed=True)  # noqa: B019
        def _get_level_name(self, record: logging.LogRecord, default_level: str = "INFO") -> str:
            """Determine the loguru record name for the given Python logging record.
            :param logging.LogRecord record: The record to determine the level name from
            :return: The logging level name for Loguru
            :rtype: str
            """

            valid_names = list(loguru.logger._core.levels.keys())  # type: ignore
            if record.levelname in valid_names:
                return record.levelname

            return default_level

        def emit(self, record: logging.LogRecord):  # pragma: no cover
            """Given a :class:`logging.LogRecord`, handle it with Loguru logging.
            :param ~logging.LogRecord record: The log record to handle
            """

            frame, depth = logging.currentframe(), 2  # type: ignore
            while frame.f_code.co_filename == logging.__file__:
                if frame.f_back is not None:
                    frame = frame.f_back
                depth += 1

            loguru.logger.opt(depth=depth, exception=record.exc_info).log(
                self._get_level_name(record), record.getMessage()
            )

    @classmethod
    def is_handled(cls, logger: Logger) -> bool:
        """Quick helper method to check if the given logger is already being handled.
        :param ~loguru._logger.Logger logger: The logger to check if already handled
        :return: True if the logger is already handled, otherwise False
        :rtype: bool
        """

        return cls._is_intercepting

    @classmethod
    def add_handle(cls, logger: Logger) -> bool:
        """Add the logging handler to the given logger instance.
        :param ~loguru._logger.Logger logger: The logger instance to add the handler to
        :return: True if the handler was added, False if the handle was already present
        :rtype: bool
        """

        if cls.is_handled(logger):
            return False

        cls._previous_handlers = copy.copy(logging._handlerList)  # type: ignore
        logging.basicConfig(handlers=[cls.LoggingHandler()], level=logging.NOTSET)
        cls._is_intercepting = True
        return True

    @classmethod
    def remove_handle(cls, logger: Logger) -> bool:
        """Remove the logging handler from the given logger instance.
        :param ~loguru._logger.Logger logger: The logger instance to remove the handle
        :return: True if the handler was removed, False if there was no handler present
        :rtype: bool
        """

        if not cls.is_handled(logger):
            return False

        logging.basicConfig(handlers=cls._previous_handlers)
        cls._is_intercepting = False
        return True



# -*- coding: utf-8 -*-

"""Loguru Client"""
import sys
from functools import lru_cache
from typing import Optional

import loguru
from loguru._logger import Logger

# Custom logger color patterns to make things a bit prettier
from lognub.captures import python_exceptions, python_warnings
from lognub.handles import PropagateHandler
from lognub.handles.intercept_handler import InterceptHandler
from lognub.patchers import patch_logger

LOGGER_LEVELS = [
    {"name": "TRACE", "color": "<fg #ffffff><bg #757575>"},
    {"name": "DEBUG", "color": "<fg #757575>"},
    {"name": "INFO", "color": "<fg #2196F3>"},
    {"name": "SUCCESS", "color": "<fg #66BB6A><bold>"},
    {"name": "WARNING", "color": "<fg #FFA000>"},
    {"name": "ERROR", "color": "<fg #E53935><bold>"},
    {"name": "CRITICAL", "color": "<fg #ffffff><bg #E53935><bold>"},
]

DEFAULT_HANDLER = dict(sink=sys.stdout, level="INFO")  # noqa: C408
DEBUG_HANDLER = dict(sink=sys.stdout, level="DEBUG", backtrace=True, diagnose=True)  # noqa: C408

# The overall logging configuration we use for our application's root logger
# https://bit.ly/2NV6al8
LOGGER_DEFAULT_CONFIG = dict(levels=LOGGER_LEVELS, handlers=[DEFAULT_HANDLER])  # noqa: C408
LOGGER_DEBUG_CONFIG = dict(levels=LOGGER_LEVELS, handlers=[DEBUG_HANDLER])  # noqa: C408


def configure_logger(
    logger_config: Optional[dict] = None,
    capture_warnings: bool = False,
    capture_exceptions: bool = False,
    propagate: bool = False,
    intercept: bool = False,
):
    """Configure the root logger based on what settings we have declared.
    For additional information on what to define in the ``logger_config`` parameter,
    checkout the `Loguru configuration <https://bit.ly/2NV6al8>`_.
    .. important:: Everytime this method is called we are cycling the added handlers and
        thus their logger ids. This is required to implement the subsequent toggling of
        features that loguru handles. Because of this, the returned loguru handler ids
        are unreliable as they may be cycled by any other module which has imported this
        method.
        *In a perfect world* we should only ever configure the logger once and only once
        at the beginning of execution and toggle features using environment variables.
        Because of this, *it is recommended to avoid calling this method at all*. The
        necessary logic to configure loguru is already provided in the
        :meth:`~modist.log.get_logger` method and can be easily accessed via
        the ``instance`` property:
        >>> log.info("This is a message")
    .. caution:: This method cannot handle both propagation and interception of logs as
        doing so would result in a recursive event where all logged propagated messages
        would be intercepted and re-propagated. In reality, you should
        **rarely if ever** be using ``intercept`` as you *almost* never want to override
        how Python's default logging works. Since so much external tooling depends on
        Python's nasty logging framework, intercepting logs is seen as a bad practice.
    :param Optional[dict] logger_config: Loguru logger configuration dictionary
    :param bool capture_warnings: Logger captures all Python warnings if set to True,
        defaults to False
    :param bool capture_exceptions: Logger captures all unhandled Python exceptions if
        set to True, defaults to False
    :param bool propagate: Logger propagates all log records to Python's builtin logging
        library if set to True, defaults to False
    :param bool intercept: Logger intercepts all log records from **previously built**
        Python loggging loggers and relogs messages to the loguru logger if set to True,
        defaults to False
    :raises ValueError: When both ``propagate`` and ``intercept`` are truthy
    """

    if propagate and intercept:
        raise ValueError(
            "Cannot both propagate and intercept Python logging at the same time, "
            "this creates a circular dependency with how logging records are handled"
        )

    if not logger_config:
        logger_config = LOGGER_DEFAULT_CONFIG

    loguru.logger.configure(**logger_config)

    if capture_warnings:
        python_warnings.capture(loguru.logger)  # type: ignore
    else:
        python_warnings.release()

    if capture_exceptions:
        python_exceptions.capture(loguru.logger)  # type: ignore
    else:
        python_exceptions.release()

    # NOTE: We are forced to ignore types against loguru.logger as it performs partial
    # and renamed imports from loguru._logger.Logger as _Logger which breaks mypy
    # analysis and really can't safely be retyped in this module

    # TODO: I would like to generalize this adding and removing of handlers since this
    # code looks a little too duplicative
    if propagate:
        PropagateHandler.add_handle(loguru.logger)  # type: ignore
    else:
        PropagateHandler.remove_handle(loguru.logger)  # type: ignore

    if intercept:
        InterceptHandler.add_handle(loguru.logger)  # type: ignore
    else:
        InterceptHandler.remove_handle(loguru.logger)  # type: ignore


@lru_cache(maxsize=64, typed=True)
def get_logger(log_level: str = "warn") -> Logger:
    """Get a logger instance that can be used right out of the box.
    >>> log = get_logger()
    >>> log.info("This is a message")
    2019-09-25 15:53:08.295 | INFO     | __master__:20 - This is a message
    :return: A logger instance
    :rtype: ~loguru._logger.Logger
    """

    # build a patched logger instance based on record patches we support in `patchers`
    patched_logger: Logger = patch_logger(loguru.logger)  # type: ignore

    # NOTE: loguru does not supply the deprecated (?) `warn()` alias for `warning()`.
    # So we just patch that alias in, users should utilize `warning` instead of `warn`.
    setattr(patched_logger, log_level, patched_logger.warning)

    configure_logger(capture_warnings=True, capture_exceptions=True, propagate=False)
    return patched_logger


"""Instantiated logger instance that you can quickly use to log anything."""
# log: Logger = get_logger()
log_instance: Logger = get_logger()
log: Logger = log_instance
# from ..log import log_instance as log



# -*- coding: utf-8 -*-
import atexit
import copy
import sys
from functools import partial
from types import TracebackType
from typing import Type

from loguru._logger import Logger

_ORIGINAL_EXCEPTHOOK = copy.deepcopy(sys.excepthook)


def _excepthook(
    logger: Logger,
    exception_type: Type[Exception],
    exception: Exception,
    traceback: TracebackType,
):
    """Substitute :func:`sys.excepthook` method.
    .. important:: This handler will and should **always** re-call the previous
        exception hook callable as we should absolutely never silently pass on all
        exceptions raised by the Python runtime. If you do want to do something stupid
        like that, you should first override what :func:`sys.excepthook` is before
        calling :func:`~capture` from here.
    :param ~loguru._logger.Logger logger: The logger to use to log the exception
    :param Type[Exception] exception_type: The type of the exception being raised
    :param Exception exception: The instance of the exception being raised
    :param TracebackType traceback: The traceback of the exception being raised
    """

    logger.exception(exception)
    _ORIGINAL_EXCEPTHOOK(exception_type, exception, traceback)


def is_captured() -> bool:
    """Check if Python exceptions are currently being captured.
    :return: True if exceptions are being captured, otherwise False
    :rtype: bool
    """

    return sys.excepthook != _ORIGINAL_EXCEPTHOOK


def release() -> bool:
    """Release the current capture of Python exceptions by loguru loggers.
    :return: True if exceptions are no longer being captured, False if they wer not
        already being captured
    :rtype: bool
    """

    if not is_captured():
        return False

    sys.excepthook = _ORIGINAL_EXCEPTHOOK
    return True


def capture(logger: Logger) -> bool:
    """Start capture of Python exceptions by a given :class:`~loguru._logger.Logger`.
    :param ~loguru._logger.Logger logger: The logger to report Python exceptions through
    :return: True if exceptions have started being captured, False if they are already
        being captured
    :rtype: bool
    """

    if is_captured():
        return False

    sys.excepthook = partial(_excepthook, logger)
    atexit.register(release)
    return True



# -*- coding: utf-8 -*-
import atexit
import copy
import warnings
from functools import partial
from typing import Optional, TextIO, Type

from loguru._logger import Logger

_ORIGINAL_SHOWWARNING = copy.deepcopy(warnings.showwarning)


def _warning_handler(
    logger: Logger,
    message: str,
    category: Type[Warning],
    filename: str,
    lineno: int,
    file: Optional[TextIO] = None,
    line: Optional[str] = None,
):
    """Substitute :func:`warnings.showwarning` method.
    .. note:: We allow the first argument to be logger by overriding the
        :func:`warnings.showwarning` method using a :class:`functools.partial` which is
        passing in the logger as the first argument.
    :param Logger logger: The logger to use to log the warning.
    :param str message: The warning message
    :param Type[~warnings.Warning] category: The category of the warning be raised
    :param str filename: The name of the file raising the warning
    :param int lineno: The line number of the file raising the warning
    :param Optional[~io.TextIO] file: The open file raising the warning, optional,
        defaults to None
    :param Optional[str] line: The line content of the open file raising the warning,
        optional, defaults to None
    """

    logger.bind(category=category, filename=filename, lineno=lineno).warning(message)
    _ORIGINAL_SHOWWARNING(message, category, filename, lineno, file, line)


def is_captured() -> bool:
    """Check if Python warnings are currently being captured.
    :return: True if warnings are captured, otherwise False
    :rtype: bool
    """

    return warnings.showwarning != _ORIGINAL_SHOWWARNING


def release() -> bool:
    """Release the current capture of Python warnings by loguru loggers.
    :returns: True if warnings are no longer being captured, False if they were not
        already being captured
    :rtype: bool
    """

    if not is_captured():
        return False

    warnings.showwarning = _ORIGINAL_SHOWWARNING
    atexit.unregister(release)
    return True


def capture(logger: Logger) -> bool:
    """Start capture of Python warnings by a given :class:`~loguru._logger.Logger`.
    :param ~loguru._logger.Logger logger: The logger to report Python warnings through
    :returns: True if warnings have started being captured, False if they are already
        being captured
    :rtype: bool
    """

    if is_captured():
        return False

    warnings.showwarning = partial(_warning_handler, logger)
    atexit.register(release)
    return True
